/*
 *	utfnfc is a library for strnfc.
 *	strcompact() converts utf-8 string to a compact form.
 *	if the input is NFD, then the output is NFC.
 *	only Latin, Greek and Japanes Kana are supported.
 *	however you can easily add data in necessity.
 *
 *	-Kenar-
 */

#include <u.h>
#include <libc.h>
#include "strcompact.h"

typedef struct NfdNfcConv NfdNfcConv;
struct NfdNfcConv {
	Rune cm;	/* combining mark */
	NfdNfc *tr; /* translating rule */
};

NfdNfcConv trans[]={
0x0300, uc_0300,
0x0301, uc_0301,
0x0302, uc_0302,
0x0303, uc_0303,
0x0304, uc_0304,
0x0306, uc_0306,
0x0307, uc_0307,
0x0308, uc_0308,
0x0309, uc_0309,
0x030a, uc_030a,
0x030b, uc_030b,
0x030c, uc_030c,
0x030f, uc_030f,
0x0311, uc_0311,
0x0313, uc_0313,
0x0314, uc_0314,
0x031b, uc_031b,
0x0323, uc_0323,
0x0324, uc_0324,
0x0325, uc_0325,
0x0326, uc_0326,
0x0327, uc_0327,
0x0328, uc_0328,
0x032d, uc_032d,
0x032e, uc_032e,
0x0330, uc_0330,
0x0331, uc_0331,
0x0342, uc_0342,
0x0345, uc_0345,
0x05b4, uc_05b4,
0x05b7, uc_05b7,
0x05b8, uc_05b8,
0x05b9, uc_05b9,
0x05bc, uc_05bc,
0x05bf, uc_05bf,
0x05c1, uc_05c1,
0x05c2, uc_05c2,
0x3099, uc_3099,
0x309a, uc_309a,
};

static Rune
reduce(Rune cm, Rune from)
{
	NfdNfcConv *c = trans;
	NfdNfc *tab;
	ulong u;

	if(cm < 0x0300)
		return 0;
	u = (cm & ~0x00ff);
	if(u != 0x0300 && u != 0x0500 && u != 0x3000)
		return 0;
	while(c->cm != 0){
		if(c->cm == cm)
			break;
		c++;
	}
	if(c->cm == 0)
		return 0;

	tab =c->tr;
	while(tab->from != 0){
		if(tab->from == from)
			return tab->to;
		tab++;
	}
	return 0;
}

/*
 * convert to compact string. (NFC in most case)
 * if input string is UTF8-NFD then strcompact will produce UTF8-NFC.
 * the usage is:
 *	m = strlen(s);
 *	p = malloc(m+UTFmax);
 *	strcompact(p,m+UTFmax,s)
 *	print("%s\n",p;
 */
int
strcompact(char *buf, int m, char *s)
{
	Rune r,t,v;
	char *p,*q,*e;
	e = s + strlen(s);
	p = buf;
	q = buf + m - UTFmax;
	s += chartorune(&t, s);
	r = t;
	while((s < e) && (p < q)){
		s += chartorune(&t, s);
		if((v = reduce(t,r)))
				r = v;
		else{
			p += runetochar(p,&r);
			r = t;
		}
	}
	p += runetochar(p,&r);
	*p = 0;
	return p - buf;
}

